"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""
import re
from typing import Dict, Tuple, Iterable, Callable, List

from .assembler import MemorySectionType, ParseContext, AssemblerDirectives
from .colors import FMT_PARSE
from .helpers import Peekable
from .tokenizer import Token, TokenType, tokenize
from .types import Program, T_ParserOpts, ProgramLoader, SimpleInstruction
from .types.exceptions import ParseException


def parse_instruction(token: Token, args: Tuple[str], context: ParseContext):
    if context.section is None:
        context.new_section(".text", MemorySectionType.Instructions)
    if context.section.type != MemorySectionType.Instructions:
        raise ParseException(
            "{} {} encountered in invalid context: {}".format(token, args, context)
        )
    ins = SimpleInstruction(
        token.value, args, context.context, context.current_address()
    )
    context.section.data.append(ins)


def parse_label(token: Token, args: Tuple[str], context: ParseContext):
    name = token.value[:-1]
    if re.match(r"^\d+$", name):
        # relative label:
        context.context.numbered_labels[name].append(context.current_address())
    else:
        if name in context.context.labels:
            print(FMT_PARSE + "Warn: Symbol {} defined twice!".format(name))
        context.add_label(name, context.current_address(), is_relative=True)


PARSERS: Dict[TokenType, Callable[[Token, Tuple[str], ParseContext], None]] = {
    TokenType.PSEUDO_OP: AssemblerDirectives.handle_instruction,
    TokenType.LABEL: parse_label,
    TokenType.INSTRUCTION_NAME: parse_instruction,
}


def parse_tokens(name: str, tokens_iter: Iterable[Token]) -> Program:
    """
    Convert a token stream into a parsed program
    :param name: the programs name
    :param tokens_iter: the programs content, tokenized
    :return: a parsed program
    """
    context = ParseContext(name)

    for token, args in composite_tokenizer(Peekable[Token](tokens_iter)):
        if token.type not in PARSERS:
            raise ParseException("Unexpected token type: {}, {}".format(token, args))
        PARSERS[token.type](token, args, context)

    return context.finalize()


def composite_tokenizer(
    tokens_iter: Iterable[Token],
) -> Iterable[Tuple[Token, Tuple[str]]]:
    """
    Convert an iterator over tokens into an iterator over tuples: (token, list(token))

    The first token ist either a pseudo_op, label, or instruction name. The token list are all remaining tokens before
    a newline is encountered
    :param tokens_iter: An iterator over tokens
    :return: An iterator over a slightly more structured representation of the tokens
    """
    tokens: Peekable[Token] = Peekable[Token](tokens_iter)

    while not tokens.is_empty():
        token = next(tokens)
        if token.type in (
            TokenType.PSEUDO_OP,
            TokenType.LABEL,
            TokenType.INSTRUCTION_NAME,
        ):
            yield token, tuple(take_arguments(tokens))


def take_arguments(tokens: Peekable[Token]) -> Iterable[str]:
    """
    Consumes (argument comma)* and yields argument.value until newline is reached
    If an argument is not followed by either a newline or a comma, a parse exception is raised
    The newline at the end is consumed
    :param tokens: A Peekable iterator over some Tokens
    """
    while True:
        if tokens.peek().type == TokenType.ARGUMENT:
            yield next(tokens).value
        elif tokens.peek().type == TokenType.NEWLINE:
            next(tokens)
            break
        elif tokens.peek().type == TokenType.COMMA:
            next(tokens)
        else:
            break

        # raise ParseException("Expected newline, instead got {}".format(tokens.peek()))


class AssemblyFileLoader(ProgramLoader):
    """
    This class loads assembly files written by hand. It understands some assembler directives and supports most
    pseudo instructions. It does very little verification of source correctness.

    It also supports numbered jump targets and properly supports local and global scope (.globl assembly directive)


    The AssemblyFileLoader loads .asm, .S and .s files by default, and acts as a weak fallback to all other filetypes.
    """

    def parse(self) -> Program:
        with open(self.source_path, "r") as f:
            return parse_tokens(self.filename, tokenize(f))

    def parse_io(self, io: Iterable[str]):
        return parse_tokens(self.filename, tokenize(io))

    @classmethod
    def can_parse(cls, source_path: str) -> float:
        """

        It also acts as a weak fallback if no other loaders want to take the file.

        :param source_path: the path to the source file
        :return:
        """
        # gcc recognizes these line endings as assembly. So we will do too.
        if source_path.split(".")[-1] in ("asm", "S", "s"):
            return 1
        return 0.01

    @classmethod
    def get_options(cls, argv: List[str]) -> Tuple[List[str], T_ParserOpts]:
        return argv, {}



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This package aims at providing an all-round usable RISC-V emulator and debugger

It contains everything needed to run assembly files, so you don't need any custom compilers or toolchains
"""

from .types.exceptions import (
    RiscemuBaseException,
    LaunchDebuggerException,
    InvalidSyscallException,
    LinkerException,
    ParseException,
    NumberFormatException,
    InvalidRegisterException,
    MemoryAccessException,
    OutOfMemoryException,
)

from .instructions import *

from .MMU import MMU
from .registers import Registers
from .syscall import SyscallInterface, Syscall
from .CPU import CPU, UserModeCPU
from .debug import launch_debug_session

from .config import RunConfig

from .parser import tokenize, parse_tokens, AssemblyFileLoader

__author__ = "Anton Lydike <Anton@Lydike.com>"
__copyright__ = "Copyright 2023 Anton Lydike"
__version__ = "2.1.1"



from .instruction_set import InstructionSet, Instruction


class RV_Debug(InstructionSet):
    def instruction_print(self, ins: Instruction):
        reg = ins.get_reg(0)
        print("register {} contains value {}".format(reg, self.regs.get(reg)))

    def instruction_print_float(self, ins: Instruction):
        reg = ins.get_reg(0)
        print("register {} contains value {}".format(reg, self.regs.get_f(reg).value))

    def instruction_print_uint(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(
                reg, self.regs.get(reg).unsigned_value
            )
        )

    def instruction_print_hex(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(reg, hex(self.regs.get(reg).value))
        )

    def instruction_print_uhex(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(
                reg, hex(self.regs.get(reg).unsigned_value)
            )
        )



from unittest import TestCase

from riscemu.tokenizer import (
    tokenize,
    print_tokens,
    Token,
    TokenType,
    NEWLINE,
    COMMA,
    split_whitespace_respecting_quotes,
)


def ins(name: str) -> Token:
    return Token(TokenType.INSTRUCTION_NAME, name)


def arg(name: str) -> Token:
    return Token(TokenType.ARGUMENT, name)


def op(name: str) -> Token:
    return Token(TokenType.PSEUDO_OP, name)


def lbl(name: str) -> Token:
    return Token(TokenType.LABEL, name)


class TestTokenizer(TestCase):
    def test_instructions(self):
        program = ["li     a0, 144", "divi   a0, a0, 12", "xori   a1, a0, 12"]
        tokens = [
            ins("li"),
            arg("a0"),
            COMMA,
            arg("144"),
            NEWLINE,
            ins("divi"),
            arg("a0"),
            COMMA,
            arg("a0"),
            COMMA,
            arg("12"),
            NEWLINE,
            ins("xori"),
            arg("a1"),
            COMMA,
            arg("a0"),
            COMMA,
            arg("12"),
            NEWLINE,
        ]
        self.assertEqual(list(tokenize(program)), tokens)

    def test_comments(self):
        parsed_res = [ins("li"), arg("a0"), COMMA, arg("144"), NEWLINE]
        for c in ("#", "//", ";"):
            lines = [c + " this is a comment", "li a0, 144"]
            self.assertEqual(list(tokenize(lines)), parsed_res)

    def test_pseudo_ins(self):
        parsed_res = [
            Token(TokenType.PSEUDO_OP, ".section"),
            Token(TokenType.ARGUMENT, ".text"),
            NEWLINE,
            Token(TokenType.PSEUDO_OP, ".type"),
            Token(TokenType.ARGUMENT, "init"),
            COMMA,
            Token(TokenType.ARGUMENT, "@function"),
            NEWLINE,
        ]
        input_program = [".section .text", ".type init, @function"]
        self.assertEqual(list(tokenize(input_program)), parsed_res)

    def test_full_program(self):
        program = """
# a hashtag comment

; semicolon comment followed by an empty line
.section .text
// double slash comment
    addi sp, sp, -32
    sw   s0, 0(ra)
section:
    sub  s0, s0, s0
"""
        tokens = [
            op(".section"),
            arg(".text"),
            NEWLINE,
            ins("addi"),
            arg("sp"),
            COMMA,
            arg("sp"),
            COMMA,
            arg("-32"),
            NEWLINE,
            ins("sw"),
            arg("s0"),
            COMMA,
            arg("ra"),
            arg("0"),
            NEWLINE,
            lbl("section:"),
            NEWLINE,
            ins("sub"),
            arg("s0"),
            COMMA,
            arg("s0"),
            COMMA,
            arg("s0"),
            NEWLINE,
        ]

        self.assertEqual(list(tokenize(program.splitlines())), tokens)

    def test_split_whitespace_respecting_quotes_single(self):
        self.assertEqual(list(split_whitespace_respecting_quotes("test")), ["test"])

    def test_split_whitespace_respecting_quotes_empty(self):
        self.assertEqual(list(split_whitespace_respecting_quotes("")), [])

    def test_split_whitespace_respecting_quotes_two_parts(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes("test 123")), ["test", "123"]
        )

    def test_split_whitespace_respecting_quotes_whole_quoted(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes("'test 123'")), ["test 123"]
        )

    def test_split_whitespace_respecting_quotes_double_quotes(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes('"test 123"')), ["test 123"]
        )

    def test_split_whitespace_respecting_quotes_quoted_then_normal(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes('"test 123" abc')),
            ["test 123", "abc"],
        )

    def test_split_whitespace_respecting_quotes_quoted_sorrounded(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes('hello "test 123" abc')),
            ["hello", "test 123", "abc"],
        )

    def test_split_whitespace_respecting_quotes_weird_spaces(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes('hello  "test 123"\tabc')),
            ["hello", "test 123", "abc"],
        )

    def test_split_whitespace_respecting_quotes_quotes_no_spaces(self):
        self.assertEqual(
            list(split_whitespace_respecting_quotes('hello"test 123"abc')),
            ["hello", "test 123", "abc"],
        )



from abc import ABC, abstractmethod
from typing import List, Type, Callable, Set, Dict, TYPE_CHECKING, Iterable

from ..registers import Registers
from ..config import RunConfig
from ..colors import FMT_NONE, FMT_CPU
from . import T_AbsoluteAddress, Instruction, Program, ProgramLoader

if TYPE_CHECKING:
    from ..MMU import MMU
    from ..instructions import InstructionSet


class CPU(ABC):
    # static cpu configuration
    INS_XLEN: int = 4

    # housekeeping variables
    regs: Registers
    mmu: "MMU"
    pc: T_AbsoluteAddress
    cycle: int
    halted: bool

    # debugging context
    debugger_active: bool

    # instruction information
    instructions: Dict[str, Callable[[Instruction], None]]
    instruction_sets: Set["InstructionSet"]

    # configuration
    conf: RunConfig

    def __init__(
        self,
        mmu: "MMU",
        instruction_sets: List[Type["InstructionSet"]],
        conf: RunConfig,
    ):
        self.mmu = mmu
        self.regs = Registers(conf.unlimited_registers)
        self.conf = conf

        self.instruction_sets = set()
        self.instructions = dict()

        for set_class in instruction_sets:
            ins_set = set_class(self)
            self.instructions.update(ins_set.load())
            self.instruction_sets.add(ins_set)

        self.halted = False
        self.cycle = 0
        self.pc = 0
        self.debugger_active = False

    def run_instruction(self, ins: Instruction):
        """
        Execute a single instruction

        :param ins: The instruction to execute
        """
        if ins.name in self.instructions:
            self.instructions[ins.name](ins)
        else:
            # this should never be reached, as unknown instructions are imparseable
            raise RuntimeError("Unknown instruction: {}".format(ins))

    def load_program(self, program: Program):
        self.mmu.load_program(program)

    def __repr__(self):
        """
        Returns a representation of the CPU and some of its state.
        """
        return "{}(pc=0x{:08X}, cycle={}, halted={} instructions={})".format(
            self.__class__.__name__,
            self.pc,
            self.cycle,
            self.halted,
            " ".join(s.name for s in self.instruction_sets),
        )

    @abstractmethod
    def step(self, verbose: bool = False):
        pass

    @abstractmethod
    def run(self, verbose: bool = False):
        pass

    def launch(self, verbose: bool = False):
        entrypoint = self.mmu.find_entrypoint()

        if entrypoint is None:
            entrypoint = self.mmu.programs[0].entrypoint

        if self.conf.verbosity > 0:
            print(
                FMT_CPU
                + "[CPU] Started running from {}".format(
                    self.mmu.translate_address(entrypoint)
                )
                + FMT_NONE
            )
        self.pc = entrypoint
        self.run(verbose)

    @classmethod
    @abstractmethod
    def get_loaders(cls) -> Iterable[Type[ProgramLoader]]:
        pass

    def get_best_loader_for(self, file_name: str) -> Type[ProgramLoader]:
        return max(self.get_loaders(), key=lambda ld: ld.can_parse(file_name))

    @property
    def sections(self):
        return self.mmu.sections

    @property
    def programs(self):
        return self.mmu.programs



from riscemu.colors import FMT_ERROR, FMT_NONE, FMT_BOLD, FMT_GREEN
from riscemu.instructions import InstructionSet
from riscemu.types import Instruction, CPU
from riscemu.decoder import RISCV_REGS

FMT_SUCCESS = FMT_GREEN + FMT_BOLD


def assert_equals(ins: Instruction, cpu: CPU):
    a, b = (get_arg_from_ins(ins, i, cpu) for i in (0, 2))
    return a == b


def assert_equals_mem(ins: Instruction, cpu: CPU):
    a, b = (get_arg_from_ins(ins, i, cpu) for i in (0, 2))
    a = cpu.mmu.read_int(a)
    return a == b


def assert_in(ins: Instruction, cpu: CPU):
    a = get_arg_from_ins(ins, 0, cpu)
    others = [get_arg_from_ins(ins, i, cpu) for i in range(2, len(ins.args))]
    return a in others


def _not(func):
    def test(ins: Instruction, cpu: CPU):
        return not func(ins, cpu)

    return test


def get_arg_from_ins(ins: Instruction, num: int, cpu: CPU):
    a = ins.args[num]
    if a in RISCV_REGS:
        return cpu.regs.get(a)
    return ins.get_imm(num)


assert_ops = {
    "==": assert_equals,
    "!=": _not(assert_equals),
    "in": assert_in,
    "not_in": _not(assert_in),
}


class Z_test(InstructionSet):
    def __init__(self, cpu: "CPU"):
        print(
            "[Test] loading testing ISA, this is only meant for running testcases and is not part of the RISC-V ISA!"
        )
        self.failed = False
        super().__init__(cpu)

    def instruction_assert(self, ins: Instruction):
        if len(ins.args) < 3:
            print(
                FMT_ERROR + "[Test] Unknown assert statement: {}".format(ins) + FMT_NONE
            )
            return
        op = ins.args[1]
        if op not in assert_ops:
            print(
                FMT_ERROR
                + "[Test] Unknown operation statement: {} in {}".format(op, ins)
                + FMT_NONE
            )
            return

        if assert_ops[op](ins, self.cpu):
            print(FMT_SUCCESS + "[TestCase] 🟢 passed assertion {}".format(ins))
        else:
            print(FMT_ERROR + "[TestCase] 🔴 failed assertion {}".format(ins))
            self.cpu.halted = True
            self.failed = True

    def instruction_fail(self, ins: Instruction):
        print(FMT_ERROR + "[TestCase] 🔴 reached fail instruction! {}".format(ins))
        self.cpu.halted = True
        self.failed = True



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This package holds all instruction sets, available to the processor
"""

from .instruction_set import InstructionSet, Instruction
from .RV32M import RV32M
from .RV32I import RV32I
from .RV32A import RV32A
from .RV32F import RV32F
from .RV_Debug import RV_Debug

InstructionSetDict = {v.__name__: v for v in [RV32I, RV32M, RV32A, RV32F, RV_Debug]}



import math

from riscemu.types import Float32

# pi encoded as a 32bit little endian float
PI_BYTES_LE = b"\xdb\x0fI@"


def test_float_serialization():
    assert Float32(PI_BYTES_LE) == Float32(math.pi)
    assert Float32(math.pi).bytes == PI_BYTES_LE


def test_random_float_ops():
    val = Float32(5)
    assert val**2 == 25
    assert val // 2 == 2
    assert val * 3 == 15
    assert val - 2 == 3
    assert val * val == 25
    assert Float32(9) ** 0.5 == 3


def test_float_from_raw_int_conversion():
    assert Float32.from_bytes(1084227584) == Float32(5.0)



from abc import ABC, abstractmethod
from typing import Tuple


class Instruction(ABC):
    name: str
    args: tuple

    @abstractmethod
    def get_imm(self, num: int) -> int:
        """
        parse and get immediate argument
        """
        pass

    @abstractmethod
    def get_imm_reg(self, num: int) -> Tuple[int, str]:
        """
        parse and get an argument imm(reg)
        """
        pass

    @abstractmethod
    def get_reg(self, num: int) -> str:
        """
        parse and get an register argument
        """
        pass

    def __repr__(self):
        return "{} {}".format(self.name, ", ".join(self.args))



from typing import Union, Tuple

from . import Instruction, T_RelativeAddress, InstructionContext
from ..helpers import parse_numeric_argument


class SimpleInstruction(Instruction):
    def __init__(
        self,
        name: str,
        args: Union[Tuple[()], Tuple[str], Tuple[str, str], Tuple[str, str, str]],
        context: InstructionContext,
        addr: T_RelativeAddress,
    ):
        self.context = context
        self.name = name
        self.args = args
        self.addr = addr

    def get_imm(self, num: int) -> int:
        resolved_label = self.context.resolve_label(self.args[num], self.addr)
        if resolved_label is None:
            return parse_numeric_argument(self.args[num])
        return resolved_label

    def get_imm_reg(self, num: int) -> Tuple[int, str]:
        return self.get_imm(num + 1), self.get_reg(num)

    def get_reg(self, num: int) -> str:
        return self.args[num]



from typing import Dict, Any
import re

# define some base type aliases so we can keep track of absolute and relative addresses
T_RelativeAddress = int
T_AbsoluteAddress = int

# parser options are just dictionaries with arbitrary values
T_ParserOpts = Dict[str, Any]

NUMBER_SYMBOL_PATTERN = re.compile(r"^\d+[fb]$")

# base classes
from .flags import MemoryFlags
from .int32 import UInt32, Int32
from .float32 import Float32
from .instruction import Instruction
from .instruction_context import InstructionContext
from .memory_section import MemorySection
from .program import Program
from .program_loader import ProgramLoader
from .cpu import CPU
from .simple_instruction import SimpleInstruction
from .instruction_memory_section import InstructionMemorySection
from .binary_data_memory_section import BinaryDataMemorySection

# exceptions
from .exceptions import (
    ParseException,
    NumberFormatException,
    MemoryAccessException,
    OutOfMemoryException,
    LinkerException,
    LaunchDebuggerException,
    RiscemuBaseException,
    InvalidRegisterException,
    InvalidAllocationException,
    InvalidSyscallException,
    UnimplementedInstruction,
    INS_NOT_IMPLEMENTED,
)



from collections import defaultdict
from typing import Dict, List, Optional

from .exceptions import ParseException
from ..types import T_AbsoluteAddress, T_RelativeAddress, NUMBER_SYMBOL_PATTERN


class InstructionContext:
    base_address: T_AbsoluteAddress
    """
    The address where the instruction block is placed
    """

    labels: Dict[str, T_RelativeAddress]
    """
    This dictionary maps all labels to their relative position of the instruction block
    """

    numbered_labels: Dict[str, List[T_RelativeAddress]]
    """
    This dictionary maps numbered labels (which can occur multiple times) to a list of (block-relative) addresses where
    the label was placed
    """

    global_symbol_dict: Dict[str, T_AbsoluteAddress]
    """
    A reference to the MMU's global symbol dictionary for access to global symbols
    """

    def __init__(self):
        self.labels = dict()
        self.numbered_labels = defaultdict(list)
        self.base_address = 0
        self.global_symbol_dict = dict()

    def resolve_label(
        self, symbol: str, address_at: Optional[T_RelativeAddress] = None
    ) -> Optional[T_AbsoluteAddress]:
        if NUMBER_SYMBOL_PATTERN.match(symbol):
            if address_at is None:
                raise ParseException(
                    "Cannot resolve relative symbol {} without an address!".format(
                        symbol
                    )
                )

            direction = symbol[-1]
            values = self.numbered_labels.get(symbol[:-1], [])
            if direction == "b":
                return max(
                    (addr + self.base_address for addr in values if addr < address_at),
                    default=None,
                )
            else:
                return min(
                    (addr + self.base_address for addr in values if addr > address_at),
                    default=None,
                )
        else:
            # if it's not a local symbol, try the globals
            if symbol not in self.labels:
                return self.global_symbol_dict.get(symbol, None)
            # otherwise return the local symbol
            return self.labels.get(symbol, None)



from riscemu.types import Int32, UInt32


def test_logical_right_shift():
    a = Int32(100)
    assert a.shift_right_logical(0) == a
    assert a.shift_right_logical(10) == 0
    assert a.shift_right_logical(1) == 100 >> 1

    a = Int32(-100)
    assert a.shift_right_logical(0) == a
    assert a.shift_right_logical(1) == 2147483598
    assert a.shift_right_logical(10) == 4194303
    assert a.shift_right_logical(31) == 1
    assert a.shift_right_logical(32) == 0



import pytest

from riscemu.registers import Registers
from riscemu.types import Float32


def test_float_regs():
    r = Registers()
    # uninitialized register is zero
    assert r.get_f("fs0") == 0
    # get/set
    val = Float32(3.14)
    r.set_f("fs0", val)
    assert r.get_f("fs0") == val


def test_unlimited_regs_works():
    r = Registers(infinite_regs=True)
    r.get("infinite")
    r.get_f("finfinite")


def test_unknown_reg_fails():
    r = Registers(infinite_regs=False)
    with pytest.raises(RuntimeError, match="Invalid register: az1"):
        r.get("az1")



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This file holds the logic for starting the emulator from the CLI
"""
import sys

from riscemu import RiscemuBaseException
from riscemu.riscemu_main import RiscemuMain

try:
    main = RiscemuMain()
    main.run_from_cli(sys.argv[1:])
    sys.exit(main.cpu.exit_code if not main.cfg.ignore_exit_code else 0)

except RiscemuBaseException as e:
    print("Error: {}".format(e.message()))
    e.print_stacktrace()

    sys.exit(-1)



"""
Laods a memory image with debug information into memory
"""

import os.path
from typing import List, Iterable

from .ElfLoader import ElfMemorySection
from .types import MemoryImageDebugInfos
from ..assembler import INSTRUCTION_SECTION_NAMES
from ..colors import FMT_NONE, FMT_PARSE
from ..helpers import get_section_base_name
from ..types import MemoryFlags, ProgramLoader, Program, T_ParserOpts


class MemoryImageLoader(ProgramLoader):
    @classmethod
    def can_parse(cls, source_path: str) -> float:
        if source_path.split(".")[-1] == "img":
            return 1
        return 0

    @classmethod
    def get_options(cls, argv: List[str]) -> [List[str], T_ParserOpts]:
        return argv, {}

    def parse(self) -> Iterable[Program]:
        if "debug" not in self.options:
            yield self.parse_no_debug()
            return

        with open(self.options.get("debug"), "r") as debug_file:
            debug_info = MemoryImageDebugInfos.load(debug_file.read())

        with open(self.source_path, "rb") as source_file:
            data: bytearray = bytearray(source_file.read())

        for name, sections in debug_info.sections.items():
            program = Program(name)

            for sec_name, (start, size) in sections.items():
                if program.base is None:
                    program.base = start

                # in_code_sec = get_section_base_name(sec_name) in INSTRUCTION_SECTION_NAMES
                program.add_section(
                    ElfMemorySection(
                        data[start : start + size],
                        sec_name,
                        program.context,
                        name,
                        start,
                        MemoryFlags(False, True),
                    )
                )

            program.context.labels.update(debug_info.symbols.get(name, dict()))
            program.global_labels.update(debug_info.globals.get(name, set()))

            yield program

    def parse_no_debug(self) -> Program:
        print(
            FMT_PARSE
            + "[MemoryImageLoader] Warning: loading memory image without debug information!"
            + FMT_NONE
        )

        with open(self.source_path, "rb") as source_file:
            data: bytes = source_file.read()

        p = Program(self.filename)
        p.add_section(
            ElfMemorySection(
                bytearray(data), ".text", p.context, p.name, 0, MemoryFlags(False, True)
            )
        )
        return p

    @classmethod
    def instantiate(cls, source_path: str, options: T_ParserOpts) -> "ProgramLoader":
        if os.path.isfile(source_path + ".dbg"):
            return MemoryImageLoader(
                source_path, dict(**options, debug=source_path + ".dbg")
            )
        return MemoryImageLoader(source_path, options)



import setuptools
from glob import glob

import riscemu

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="riscemu",
    version=riscemu.__version__,
    author=riscemu.__author__,
    author_email="pip@antonlydike.de",
    description="RISC-V userspace and machine mode emulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/antonlydike/riscemu",
    project_urls={
        "Bug Tracker": "https://github.com/AntonLydike/riscemu/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=[
        "riscemu",
        "riscemu.decoder",
        "riscemu.instructions",
        "riscemu.IO",
        "riscemu.priv",
        "riscemu.types",
    ],
    package_data={
        "riscemu": ["libc/*.s", "py.typed"],
    },
    data_files=[
        ("libc", glob("libc/*.s")),
    ],
    scripts=["riscemu/tools/riscemu"],
    python_requires=">=3.8",
    install_requires=["pyelftools~=0.27"],
)



import struct
from ctypes import c_float
from typing import Union, Any

bytes_t = bytes


class Float32:
    __slots__ = ("_val",)

    _val: c_float

    @property
    def value(self) -> float:
        """
        The value represented by this float
        """
        return self._val.value

    @property
    def bytes(self) -> bytes:
        """
        The values bit representation (as a bytes object)
        """
        return struct.pack("<f", self.value)

    @property
    def bits(self) -> int:
        """
        The values bit representation as an int (for easy bit manipulation)
        """
        return int.from_bytes(self.bytes, byteorder="little")

    @classmethod
    def from_bytes(cls, val: Union[int, bytes_t, bytearray]):
        if isinstance(val, int):
            val = struct.unpack("!f", struct.pack("!I", val))[0]
        return Float32(val)

    def __init__(
        self, val: Union[float, c_float, "Float32", bytes_t, bytearray, int] = 0
    ):
        if isinstance(val, (float, int)):
            self._val = c_float(val)
        elif isinstance(val, c_float):
            self._val = c_float(val.value)
        elif isinstance(val, (bytes, bytearray)):
            self._val = c_float(struct.unpack("<f", val)[0])
        elif isinstance(val, Float32):
            self._val = val._val
        else:
            raise ValueError(
                "Unsupported value passed to Float32: {} ({})".format(
                    repr(val), type(val)
                )
            )

    def __add__(self, other: Union["Float32", float]):
        if isinstance(other, Float32):
            other = other.value
        return self.__class__(self.value + other)

    def __sub__(self, other: Union["Float32", float]):
        if isinstance(other, Float32):
            other = other.value
        return self.__class__(self.value - other)

    def __mul__(self, other: Union["Float32", float]):
        if isinstance(other, Float32):
            other = other.value
        return self.__class__(self.value * other)

    def __truediv__(self, other: Any):
        return self // other

    def __floordiv__(self, other: Any):
        if isinstance(other, Float32):
            other = other.value
        return self.__class__(self.value // other)

    def __mod__(self, other: Union["Float32", float]):
        if isinstance(other, Float32):
            other = other.value
        return self.__class__(self.value % other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (float, int)):
            return self.value == other
        elif isinstance(other, Float32):
            return self.value == other.value
        return False

    def __neg__(self):
        return self.__class__(-self.value)

    def __abs__(self):
        return self.__class__(abs(self.value))

    def __bytes__(self) -> bytes_t:
        return self.bytes

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.value)

    def __str__(self):
        return str(self.value)

    def __format__(self, format_spec: str):
        return self.value.__format__(format_spec)

    def __hash__(self):
        return hash(self.value)

    def __gt__(self, other: Any):
        if isinstance(other, Float32):
            other = other.value
        return self.value > other

    def __lt__(self, other: Any):
        if isinstance(other, Float32):
            other = other.value
        return self.value < other

    def __le__(self, other: Any):
        if isinstance(other, Float32):
            other = other.value
        return self.value <= other

    def __ge__(self, other: Any):
        if isinstance(other, Float32):
            other = other.value
        return self.value >= other

    def __bool__(self):
        return bool(self.value)

    def __pow__(self, power, modulo=None):
        if modulo is not None:
            raise ValueError("Float32 pow with modulo unsupported")
        return self.__class__(self.value**power)

    # right handed binary operators

    def __radd__(self, other: Any):
        return self + other

    def __rsub__(self, other: Any):
        return self.__class__(other) - self

    def __rmul__(self, other: Any):
        return self * other

    def __rtruediv__(self, other: Any):
        return self.__class__(other) // self

    def __rfloordiv__(self, other: Any):
        return self.__class__(other) // self

    def __rmod__(self, other: Any):
        return self.__class__(other) % self

    def __rand__(self, other: Any):
        return self.__class__(other) & self

    def __ror__(self, other: Any):
        return self.__class__(other) | self

    def __rxor__(self, other: Any):
        return self.__class__(other) ^ self

    # bytewise operators:

    def __and__(self, other: Union["Float32", float, int]):
        if isinstance(other, float):
            other = Float32(other)
        if isinstance(other, Float32):
            other = other.bits
        return self.from_bytes(self.bits & other)

    def __or__(self, other: Union["Float32", float]):
        if isinstance(other, float):
            other = Float32(other)
        if isinstance(other, Float32):
            other = other.bits
        return self.from_bytes(self.bits | other)

    def __xor__(self, other: Union["Float32", float]):
        if isinstance(other, float):
            other = Float32(other)
        if isinstance(other, Float32):
            other = other.bits
        return self.from_bytes(self.bits ^ other)

    def __lshift__(self, other: int):
        return self.from_bytes(self.bits << other)

    def __rshift__(self, other: int):
        return self.from_bytes(self.bits >> other)



import os
from abc import abstractmethod, ABC
from typing import Union, Iterator, List

from . import T_ParserOpts, Program


class ProgramLoader(ABC):
    """
    A program loader is always specific to a given source file. It is a place to store all state
    concerning the parsing and loading of that specific source file, including options.
    """

    def __init__(self, source_path: str, options: T_ParserOpts):
        self.source_path = source_path
        self.options = options
        self.filename = os.path.split(self.source_path)[-1]

    @classmethod
    @abstractmethod
    def can_parse(cls, source_path: str) -> float:
        """
        Return confidence that the file located at source_path
        should be parsed and loaded by this loader
        :param source_path: the path of the source file
        :return: the confidence that this file belongs to this parser
        """
        pass

    @classmethod
    @abstractmethod
    def get_options(cls, argv: List[str]) -> [List[str], T_ParserOpts]:
        """
        parse command line args into an options dictionary

        :param argv: the command line args list
        :return: all remaining command line args and the parser options object
        """
        pass

    @classmethod
    def instantiate(cls, source_path: str, options: T_ParserOpts) -> "ProgramLoader":
        """
        Instantiate a loader for the given source file with the required arguments

        :param source_path: the path to the source file
        :param options: the parsed options (guaranteed to come from this classes get_options method.
        :return: An instance of a ProgramLoader for the spcified source
        """
        return cls(source_path, options)

    @abstractmethod
    def parse(self) -> Union[Program, Iterator[Program]]:
        """

        :return:
        """
        pass



from riscemu.helpers import *


def test_align_address():
    assert align_addr(3, 1) == 3
    assert align_addr(3, 2) == 4
    assert align_addr(3, 4) == 4
    assert align_addr(3, 8) == 8
    assert align_addr(8, 8) == 8


def test_parse_numeric():
    assert parse_numeric_argument("13") == 13
    assert parse_numeric_argument("0x100") == 256
    assert parse_numeric_argument("-13") == -13
